"""
test_model.py
The MLOps framework needs `ServiceTest` to test services locally and make
test calls after deploying services.
It's highly recommended to define tests for training, predictions and data
validation, but it's not required by the framework.
"""
import os
from typing import Any, List

import pytest

from model_code.model import ModelIO, initialization, predict


class ServiceTest:
    """
    MLOps framework uses this object to test services
    Training-related methods and attributes are not required by the framework
    if there's no training service.
    """

    def __init__(self):
        """
        Define known inputs: `training_input` and `prediction_input`.
        Note that the values need to be customized to your model.
        """
        self.prediction_input = {
            "input": [
                {
                    "well": "dummy",
                    "input_logs": {
                        "ACS": [-1],
                        "RDEP": [1],
                        "DEN": [1],
                    },
                }
            ],
            "return_file": False,
        }

    def prediction_check(self, y: List):
        """
        Evaluate the prediction service's output when the input is
        `self.prediction_input`
        Input:
        - y: output generated by `predict`
        Output:
        - `True` if `y` is correct, `False` otherwise.
        """
        assert y == [1, 1, 1]
        return True


################################################################################
# The rest of the file are tests for the example model. It's recommended that
# you include tests to check your code. They aren't required by the framework,
# but if they're present they will be run by pytest during deployment
################################################################################

# We will reuse the inputs and checks in the unit tests. Not required though.
service_test = ServiceTest()  # type: ignore


def _check_model(model: Any) -> None:
    x = service_test.prediction_input
    y = predict(x, model)
    assert service_test.prediction_check(y)


def test_model_loading_error():
    with pytest.raises(ModelIO):
        initialization(folder_path="path_doesnt_exist")


def test_predict():
    model = initialization("model_artifact")  # model_storage[0]
    _check_model(model)


def test_predict_local_model():
    """
    Test prediction service using a model artifact to load the model.
    This will fail in the training service, since model artifacts are not
    deployed in this case. It's important to skip this test as shown below:
    """
    if os.getenv("SERVICE_NAME") == "training":
        pytest.skip("This test is specific to prediction services")
    model = initialization("model_artifact")  # Same as in settings file!
    _check_model(model)



import os
import subprocess
from pathlib import Path
from akerbp.mlops.core import config
from akerbp.mlops.core.exceptions import VirtualEnvironmentError
from akerbp.mlops.deployment.helpers import (
    create_venv,
    install_requirements,
    delete_venv,
    set_up_requirements,
)

test_venv = "test-venv"
requirements_file = "test/test_requirements.txt"


def test_create_venv_creates_venv_no_exception_raised(venv_name=test_venv):
    try:
        create_venv(venv_name=venv_name)
    except VirtualEnvironmentError as exc:
        raise AssertionError(f"Virtual environment not created ({exc})") from None


def test_install_requirements_install_single_test_requirement_in_venv_with_no_deps(
    req_file=requirements_file, venv_name=test_venv
):
    venv_dir = os.path.join(os.getcwd(), venv_name)
    install_requirements(
        req_file=req_file, venv_dir=venv_dir, with_deps=False, setup_venv=True
    )
    cmd = ["bin/pip", "list"]
    output = subprocess.Popen(
        cmd, cwd=venv_dir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = output.communicate()
    package, version = stdout.decode().split("\n")[2:-1][0].split()
    expected = {
        "package": "minimal",
        "version": "0.1.0",
    }

    assert package == expected["package"] and version == expected["version"]


def test_delete_venv_deletes_venv_no_exception_raised(venv_name=test_venv):
    try:
        delete_venv(venv_name=venv_name)
    except VirtualEnvironmentError as exc:
        raise AssertionError(f"Virtual environment not deleted ({exc})") from None


def test_set_up_requirements_from_settings_install_deps_in_venv():
    project_settings = config.read_project_settings(
        yaml_file=Path("test/test_settings/installing_reqs_in_venv.yaml")
    )
    deps_installed_in_venv = []
    expected = {
        "package": "minimal",
        "version": "0.1.0",
    }
    for setting in project_settings:
        venv_dir = set_up_requirements(
            c=setting,
            venv_name=test_venv,
            with_deps=False,
            unit_testing=True,
            setup_venv=True,
        )
        cmd = ["bin/pip", "list"]
        output = subprocess.Popen(
            cmd, cwd=venv_dir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        stdout, _ = output.communicate()
        package, version = stdout.decode().split("\n")[2:-1][0].split()

        if package == expected["package"] and version == expected["version"]:
            deps_installed_in_venv.append(True)
        else:
            deps_installed_in_venv.append(False)
    delete_venv(venv_name=test_venv)
    assert sum(deps_installed_in_venv) == len(deps_installed_in_venv)



import os

import akerbp.mlops.model_manager as mm
from akerbp.mlops.cdf.helpers import delete_file

mm.setup()
env = os.environ["MODEL_ENV"]
model_name = "mlopsdemo"
version = mm.get_latest_model_in_env(model_name, env)
delete_file(id={"external_id": f"{model_name}/{env}/{version}"})



import os

from increment_package_version import get_current_version, increment_version


def test_get_current_version() -> None:
    _ = get_current_version()


def test_get_current_version_return_zero_pre_release_number():
    version = increment_version(version_to_test="0.1.0", commit_msg="prerelease")
    pre_release_number = version.split("a")[1]
    assert pre_release_number == "0"


def test_get_current_version_return_incremented_pre_release_number():
    version = increment_version(version_to_test="0.1.0a0", commit_msg="prerelease")
    pre_release_number = version.split("a")[1]
    assert pre_release_number == "1"


def test_increment_version_major_increment_in_test():
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    expected_new_version = f"{major+1}.0.0a0"
    commit_msg = "This is a major update with breaking changes, in test environment"
    new_version = increment_version(version_to_test=version, commit_msg=commit_msg)
    assert new_version == expected_new_version


def test_increment_version_minor_increment_in_test():
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    expected_new_version = f"{major}.{int(minor)+1}.0a0"
    commit_msg = "This is a minor update without breaking changes, in test environment"
    new_version = increment_version(version_to_test=version, commit_msg=commit_msg)
    assert new_version == expected_new_version


def test_increment_version_patch_increment_in_test():
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    expected_new_version = f"{major}.{minor}.{int(patch)+1}a0"
    commit_msg = "This is a patch with a bugfix"
    new_version = increment_version(version_to_test=version, commit_msg=commit_msg)
    assert new_version == expected_new_version


def test_increment_version_major_increment_in_prod():
    os.environ["MODEL_ENV"] = "test"
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    commit_msg = "This is a major update with breaking changes, in prod environment"
    incremented_version = increment_version(
        version_to_test=version, commit_msg=commit_msg
    )
    os.environ["MODEL_ENV"] = "prod"
    new_version_in_prod = increment_version(version_to_test=incremented_version)
    expected_new_version = f"{major+1}.0.0"
    assert new_version_in_prod == expected_new_version


def test_increment_version_minor_increment_in_prod():
    os.environ["MODEL_ENV"] = "test"
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    commit_msg = "This is a minor update with breaking changes, in prod environment"
    incremented_version = increment_version(
        version_to_test=version, commit_msg=commit_msg
    )
    os.environ["MODEL_ENV"] = "prod"
    new_version_in_prod = increment_version(version_to_test=incremented_version)
    expected_new_version = f"{major}.{minor+1}.0"
    assert new_version_in_prod == expected_new_version


def test_increment_version_patch_increment_in_prod():
    os.environ["MODEL_ENV"] = "test"
    version = "0.1.0"
    major, minor, patch = list(map(int, version.split(".")))
    commit_msg = "This is just a patch in prod"
    incremented_version = increment_version(
        version_to_test=version, commit_msg=commit_msg
    )
    os.environ["MODEL_ENV"] = "prod"
    new_version_in_prod = increment_version(version_to_test=incremented_version)
    expected_new_version = f"{major}.{minor}.{patch+1}"
    assert new_version_in_prod == expected_new_version



from pathlib import Path
from typing import List

from akerbp.mlops.core.config import read_project_settings
from akerbp.mlops.core.helpers import input_data_validation

required_input = ["WELL", "DEN", "AC", "NEU", "PEF"]


def test_input_data_validation_correct_input_logs_return_true(
    required_input: List = required_input,
):
    input_with_logs = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_logs
    )
    assert validated


def test_input_data_validation_correct_input_logs_not_standard_return_true(
    required_input: List = required_input,
):
    input_with_logs = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "WELL": [],
                    "RHOB": [],
                    "DTC": [],
                    "NPHI": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_logs
    )
    assert validated


def test_input_data_validation_correct_input_logs_multiple_wells_return_true(
    required_input: List = required_input,
):
    input_with_multiple_logs = {
        "input": [
            {
                "well": "well1",
                "input_logs": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
            {
                "well": "well2",
                "input_logs": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_multiple_logs
    )
    assert validated


def test_input_data_validation_wrong_logs_return_false(
    required_input: List = required_input,
):
    input_with_wrong_logs = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "WELL": [],
                    "ACS": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_wrong_logs
    )
    assert not validated


def test_input_data_validation_wrong_logs_multiple_wells_return_false(
    required_input: List = required_input,
):
    input_with_wrong_logs_multiple_wells = {
        "input": [
            {
                "well": "well1",
                "input_logs": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
            {
                "well": "well2",
                "input_logs": {
                    "WELL": [],
                    "ACS": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_wrong_logs_multiple_wells
    )
    assert not validated


def test_input_data_validation_wrongly_specified_input_logs_return_false(
    required_input: List = required_input,
):
    input_wrongly_specified_input_logs = {
        "input": [
            {
                "well": "well-123",
                "required_input": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_wrongly_specified_input_logs
    )
    assert not validated


def test_input_data_validation_wrongly_specified_input_logs_multiple_wells_return_false(
    required_input: List = required_input,
):
    input_wrongly_specidied_input_logs_multiple_wells = {
        "input": [
            {
                "well": "well-1",
                "input_curves": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
            {
                "well": "well-2",
                "required_input_curves": {
                    "WELL": [],
                    "DEN": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    validated = input_data_validation(
        required_input=required_input,
        input=input_wrongly_specidied_input_logs_multiple_wells,
    )
    assert not validated


def test_input_data_validation_with_well_and_kwargs_only_return_true(
    required_input: List = required_input,
):
    input_with_well_and_kwargs_only = {
        "input": [{"well": "well-123", "keyword_arguments": {}}]
    }
    validated = input_data_validation(
        required_input=required_input, input=input_with_well_and_kwargs_only
    )
    assert validated


def test_input_data_validation_with_well_only_return_true(
    required_input: List = required_input,
):
    input_with_well_only = {"input": [{"well": "well-123"}]}
    validated = input_data_validation(
        required_input=required_input, input=input_with_well_only
    )
    assert validated


def test_input_data_validation_no_required_input_read_from_settings_file_only_well_in_payload(
    yaml_settings=Path("test/test_settings/single_model_no_required_input.yaml"),
):
    input_with_well_only = {"input": [{"well": "well-123"}]}
    no_required_input = []
    project_settings = read_project_settings(yaml_file=yaml_settings)
    for setting in project_settings:
        required_input = eval(setting.info["prediction"]["metadata"]["required_input"])
        validated = input_data_validation(
            required_input=required_input, input=input_with_well_only
        )
        if validated:
            no_required_input.append(True)
        else:
            no_required_input.append(False)
    assert sum(no_required_input) == len(no_required_input)


def test_input_data_validation_no_required_input_read_from_settings_file_logs_in_payload(
    yaml_settings=Path("test/test_settings/single_model_no_required_input.yaml"),
):
    input_with_logs = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "WELL": [],
                    "ACS": [],
                    "AC": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    no_required_input = []
    project_settings = read_project_settings(yaml_file=yaml_settings)
    for setting in project_settings:
        required_input = eval(setting.info["prediction"]["metadata"]["required_input"])
        validated = input_data_validation(
            required_input=required_input, input=input_with_logs
        )
        if validated:
            no_required_input.append(True)
        else:
            no_required_input.append(False)
    assert sum(no_required_input) == len(no_required_input)


def test_input_data_validation_AC_DEN_NEU_required_input_read_from_settings_file_logs_in_payload_returns_true(
    yaml_settings=Path("test/test_settings/single_model_required_input.yaml"),
):
    payload_with_AC_DEN_NEU = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "AC": [],
                    "DEN": [],
                    "NEU": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    required_input_in_payload = []
    project_settings = read_project_settings(yaml_file=yaml_settings)
    for setting in project_settings:
        required_input = eval(setting.info["prediction"]["metadata"]["required_input"])
        validated = input_data_validation(
            required_input=required_input, input=payload_with_AC_DEN_NEU
        )
        if validated:
            required_input_in_payload.append(True)
        else:
            required_input_in_payload.append(False)
    assert sum(required_input_in_payload) == len(required_input_in_payload)


def test_input_data_validation_AC_DEN_NEU_required_input_read_from_settings_file_missing_logs_in_payload_returns_false(
    yaml_settings=Path("test/test_settings/single_model_required_input.yaml"),
):
    payload_with_AC_DEN = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "AC": [],
                    "DEN": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    required_input_in_payload = []
    project_settings = read_project_settings(yaml_file=yaml_settings)
    for setting in project_settings:
        required_input = eval(setting.info["prediction"]["metadata"]["required_input"])
        validated = input_data_validation(
            required_input=required_input, input=payload_with_AC_DEN
        )
        if validated:
            required_input_in_payload.append(True)
        else:
            required_input_in_payload.append(False)
    assert sum(required_input_in_payload) < len(required_input_in_payload)


def test_input_data_validation_AC_DEN_NEU_required_input_read_from_settings_file_additional_logs_in_payload_returns_true(
    yaml_settings=Path("test/test_settings/single_model_required_input.yaml"),
):
    payload_with_additional_curves = {
        "input": [
            {
                "well": "well-123",
                "input_logs": {
                    "AC": [],
                    "DEN": [],
                    "NEU": [],
                    "PEF": [],
                },
                "keyword_arguments": {},
            },
        ]
    }
    required_input_in_payload = []
    project_settings = read_project_settings(yaml_file=yaml_settings)
    for setting in project_settings:
        required_input = eval(setting.info["prediction"]["metadata"]["required_input"])
        validated = input_data_validation(
            required_input=required_input, input=payload_with_additional_curves
        )
        if validated:
            required_input_in_payload.append(True)
        else:
            required_input_in_payload.append(False)
    assert sum(required_input_in_payload) == len(required_input_in_payload)



"""
model.py
This is a model template that shows the interface expected by MLOps framework
for an example model (scikit-learn's DummyClassifier).

Internal elements start by "_", as usually done in Python. This points to parts
of this file that are specific to the example and not required by the framework.

We will develop a single service, namely prediction/inference. Refer to the history for code needed to include a training service (was never used)
"""
import os
from typing import Any, Dict, List

from joblib import load

from akerbp.mlops.core.logger import get_logger

logger = get_logger(__name__)

_model_name = "dummy.joblib"


class ModelException(Exception):
    """
    This exception should be raised when there are errors (required by the
    framework). It can be subclassed if more specific exceptions are needed, but
    this isn't required by the framework.
    """

    pass


class ModelIO(ModelException):
    pass


class ModelDataValidation(ModelException):
    pass


def _load_model(folder_path: str) -> Any:
    """
    Load model artifact generated by `train`
    Input:
      - folder path (string) that contains model artifacts
    Output:
      - clf: model object required by `predict`. If there are multiple objects,
      you can return a tuple e.g. `(model_1, model_2)`
    """
    source_path = os.path.join(folder_path, _model_name)
    try:
        clf = load(source_path)
    except Exception as error:
        error_type = type(error).__name__
        error_message = f"Could not load model. {error_type}: {error}"
        logger.error(error_message)
        raise ModelIO(error_message) from error
    logger.info(f"Read model {clf=} from {source_path=}")
    return clf


def initialization(folder_path: str, secrets=None) -> Any:
    """
    Initialize objects required by `predict`.
    This can involve e.g. loading model artifacts generated by `train`,
    instantiate objects required to read data, etc.

    Input:
      - folder path (string) that contains model artifacts
      - secrets: dictionary with api keys, keys are "files" and "data"

    Output: object required by predict. If there are multiple objects,
      this could be a tuple, e.g. `(model, client)`
    """
    model = _load_model(folder_path)
    return model


def _validate_input(data: Dict):
    """
    Validate features sent by prediction or training service API.
    Generate ModelDataValidation with a useful message for the user.
    """
    if not isinstance(data, dict):
        raise ModelDataValidation("Data should be a dictionary")
    if "input" not in data.keys():
        raise ModelDataValidation("Input dictionary should have a column 'input'")
    if not isinstance(data["input"], list):
        raise ModelDataValidation("field 'input' should contain an array")


def _process(data_dict: Dict) -> List:
    """
    Transform data provided by the prediction service API into the data
    expected by the model.
    """
    input_dict = data_dict["input"][0]["input_logs"]
    return [input_dict[key][0] for key in input_dict.keys()]


def predict(data_dict, init_object, secrets=None) -> Any:
    """
    Predict data provided through the model service API

    Inputs:
      - data_dict: data (dictionary)
      - init_object: object provided by `load_model`. If `initialization` returns
          a tuple, you can unpack it, e.g. `model, client = init_object`
      - secrets: dictionary with api keys, keys are "files" and "data"
    Output:
      - y: prediction provided by `model` for `x`
        Note: prediction should be jsonifyable, which means for example that any
        np array should be converted to list
    """
    model = init_object
    _validate_input(data_dict)
    x = _process(data_dict)
    y = model.predict(x)
    logger.info(f"Prediction for {x=} was {y=}")
    prediction = y.tolist()
    return prediction



import os

import akerbp.mlops.model_manager as mm

model_name = "mlopsdemo"
folder_path = "model_artifact"
env = os.environ["MODEL_ENV"]
metadata = {"Description": "Dummy artifacts for mlops demo model"}

mm.setup()
folder_info = mm.upload_new_model_version(model_name, env, folder_path, metadata)



"""
deploy.py

Deploy services in either Google Cloud Run or CDF Functions.
Model registry uses CDF Files.
"""
import os
import traceback
from pathlib import Path
from typing import List

import akerbp.mlops.model_manager as mm
from akerbp.mlops import __version__ as package_version
from akerbp.mlops.core import config
from akerbp.mlops.core.config import ServiceSettings

# Global variables
from akerbp.mlops.core.logger import get_logger
from akerbp.mlops.deployment import helpers

logger = get_logger(__name__)


def deploy_model(
    model_settings: ServiceSettings,
) -> str:
    """
    Deploy a model.

    This will create a deployment folder and change current working directory
    to it.

    Return "OK" if deployment was successful, otherwise return a string with the traceback for the failed deployment

    Args:
        model_settings (ServiceSettings): settings for the model service
        platform_methods (dict): where key is the platform and value is a tuple with deploy and test functions.
            Defaults to the globally set platform_methods variable.

    Returns:
        (str): status of the deployment
    """
    c = model_settings
    try:
        original_req_file = c.req_file
        envs = config.ENV_VARS
        env = envs.model_env
        testing_only = eval(envs.testing_only)
        local_deployment = eval(envs.local_deployment)
        service_name = envs.service_name
        platform = envs.platform
        if service_name is None:
            raise ValueError(
                "SERVICE_NAME environment variable must be set to either 'prediction' or 'training'"
            )
        deployment_folder = helpers.deployment_folder_path(c.model_name)
        function_name = f"{c.model_name}-{service_name}-{env}"

        # Handle mlops internal testing settings
        mlops_testing = os.environ.get("LOCAL_MLOPS_TESTING", False)

        logger.info(
            f"Starting deployment and/or testing of model {c.human_friendly_model_name}"
        )

        if (service_name == "prediction") and c.artifact_folder:
            if platform == "cdf" or platform == "local":
                mm.set_active_dataset(c.dataset)
                c.model_id = mm.set_up_model_artifact(c.artifact_folder, c.model_name)

        logger.info("Create deployment folder and move required files/folders")
        deployment_folder.mkdir()
        # Write settings to deployment folder
        config.store_service_settings(
            c, deployment_folder / "mlops_service_settings.yaml"
        )
        logger.info("Service settings %s", c)
        # Update requirements file with mlops version
        req_file = helpers.create_temporary_copy(c.req_file)
        c.req_file = req_file

        helpers.set_mlops_import(
            req_file=c.req_file,
            platform=c.platform,
            deployment_folder=deployment_folder if mlops_testing else None,
        )
        helpers.copy_to_deployment_folder(c.files, deployment_folder)

        logger.info(f"cd {deployment_folder}")
        os.chdir(deployment_folder)

        # Local testing - for MODEL_ENV=dev and LOCAL_DEPLOYMENT=False
        if envs.model_env == "dev" and envs.local_deployment == "False":
            c.platform = "local"
        if c.platform == "local":
            if c.test_file:
                logger.info("Performing local testing")
                helpers.run_tests(
                    c,
                    setup_venv=True,
                    deploy=False,
                )
            else:
                logger.warning(
                    "No test file specified in the settings, skipping local tests."
                )

        # Deploy to CDF
        elif c.platform == "cdf":
            if (env == "test" or env == "prod") and testing_only:

                logger.info(
                    f"Running tests for model {c.human_friendly_model_name} in {env}, will not deploy"
                )
                helpers.run_tests(c, setup_venv=True, deploy=False)
            elif (
                (env == "test" or env == "prod") and not testing_only
            ) or local_deployment:
                helpers.do_deploy(
                    c,
                    env,
                    service_name,
                    function_name,
                    deployment_folder,
                    setup_venv=True,
                )
            else:
                logger.warning(
                    f"Will not run tests nor deploy model(s), check your environment variables: {config.ENV_VARS}"
                )
        return "OK"
    except (Exception, KeyboardInterrupt):
        # Remove deployment folder if it exists
        helpers.rm_deployment_folder(c.model_name)
        # Remove temporary requirements file if it exists
        if c.req_file != original_req_file:
            req_file = Path(c.req_file).resolve()
            logger.info("Removing temporary requirements file %s", req_file)
            if req_file.exists():
                req_file.unlink()
        trace = traceback.format_exc()
        return f"Model failed to deploy and/or tests failed! See the following traceback for more info: \n\n{trace}"


def deploy(project_settings: List[ServiceSettings]) -> None:
    """
    Deploy a machine learning project that potentially contains multiple models.
    Deploy each model in the settings and make sure that if one model fails it
    does not affect the rest. At the end, if any model failed, it raises an
    exception with a summary of all models that failed.

    Args:
        Project settings as described by the user in the config file.

    Raises:
        Exception: If any model failed to deploy.
    """
    failed_models = {}
    cwd_path = Path.cwd()

    for c in project_settings:
        status = deploy_model(c)
        if status != "OK":
            logger.error(status)
            failed_models[c.human_friendly_model_name] = status

        logger.info("cd ..")
        os.chdir(cwd_path)
        helpers.rm_deployment_folder(c.model_name)

    if failed_models:
        for model, message in failed_models.items():
            logger.warning(f"Model {model} failed: {message}")
        raise Exception("At least one model failed.")


def main() -> None:
    logger = get_logger(name="akerbp.mlops.deployment.deploy.py")
    logger.info(
        f"Deploying prediction service using MLOps framework version {package_version}"
    )
    mm.setup()
    settings = config.read_project_settings()
    deploy(settings)


if __name__ == "__main__":
    main()



import pytest
from pathlib import Path
from akerbp.mlops.core.config import read_project_settings
from akerbp.mlops.core.exceptions import MissingMetadataError


def test_single_model_all_required_fields_no_exception_raised():
    yaml_file = Path("test/test_settings/single_model_all_required_fields.yaml")
    try:
        read_project_settings(yaml_file=yaml_file)
    except MissingMetadataError as exc:
        raise AssertionError(f"Settings file raised an exception {exc}") from None


def test_multiple_models_all_required_fields_no_exception_raised():
    yaml_file = Path("test/test_settings/multiple_models_all_required_fields.yaml")
    try:
        read_project_settings(yaml_file=yaml_file)
    except MissingMetadataError as exc:
        raise AssertionError(f"Settings file raised an exception {exc}") from None


def test_single_model_missing_field_throws_exception():
    yaml_file = Path("test/test_settings/single_model_missing_field.yaml")
    with pytest.raises(MissingMetadataError):
        read_project_settings(yaml_file=yaml_file)


def test_single_model_missing_petrel_field_throws_exception():
    yaml_file = Path("test/test_settings/single_model_missing_petrel_field.yaml")
    with pytest.raises(MissingMetadataError):
        read_project_settings(yaml_file=yaml_file)


def test_multiple_models_one_with_missing_field_throws_exception():
    yaml_file = Path("test/test_settings/multiple_models_missing_field.yaml")
    with pytest.raises(MissingMetadataError):
        read_project_settings(yaml_file=yaml_file)



# helpers.py
import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
import venv
from importlib import resources as importlib_resources
from pathlib import Path
from typing import Any, Dict, Optional

import akerbp.mlops as akerbp_mlops
from akerbp.mlops.cdf import helpers as cdf
from akerbp.mlops.core.config import ENV_VARS, ServiceSettings
from akerbp.mlops.core.exceptions import (
    DeploymentError,
    TestError,
    VirtualEnvironmentError,
)
from akerbp.mlops.core.helpers import subprocess_wrapper
from akerbp.mlops.core.logger import get_logger
from akerbp.mlops.deployment import platforms

logger = get_logger(__name__)

envs = ENV_VARS
env = envs.model_env
service_name = envs.service_name
platform_methods = platforms.get_methods()


def is_unix() -> bool:
    """Checks whether the working OS is unix-based or not

    Returns:
        bool: True if os is unix-based
    """
    return os.name == "posix"


def get_repo_origin() -> str:
    """Get origin of the git repo

    Returns:
        (str): origin
    """
    origin = subprocess.check_output(
        ["git", "remote", "get-url", "--push", "origin"], encoding="UTF-8"
    ).rstrip()
    return origin


def replace_string_file(s_old: str, s_new: str, file: Path) -> None:
    """
    Replaces all occurrences of s_old with s_new in a specifyied file

    Args:
        s_old (str): old string
        s_new (str): new string
        file (Path): file to replace the string in
    """
    with file.open("r+") as f:
        data = f.read()
        if s_old not in data:
            logger.warning(f"Didn't find '{s_old}' in {file}")
        data = data.replace(s_old, s_new)
        new = os.linesep.join([s for s in data.splitlines() if s.strip()])
        f.seek(0)
        f.write(new)
        f.truncate()


def set_mlops_import(
    req_file: Path, platform: str, deployment_folder: Optional[str] = None
) -> None:
    """Set correct package version in requirements.txt

    Args:
        req_file (Path): path to requirements.txt for the model to deploy
        deployment_folder (Optional[str], optional): path to a deployment folder. Defaults to None.
            If this is set, the function assumes that the current version of the
            mlops package should be built and the version should be set to a wheel path.
    """
    if deployment_folder is not None:
        package_version = build_mlops_wheel(deployment_folder)
        set_to = f"./{package_version}[{platform}]"
        replace_string_file(
            "akerbp.mlops==MLOPS_VERSION",
            set_to,
            req_file,
        )
        logger.info(
            f"Set akerbp.mlops dependency to {package_version} in requirements.txt"
        )
    else:
        package_version = akerbp_mlops.__version__
        set_to = f"akerbp.mlops[{platform}]=={package_version}"
        replace_string_file(
            "akerbp.mlops==MLOPS_VERSION",
            set_to,
            req_file,
        )
        logger.info(f"Set akerbp.mlops==MLOPS_VERSION to {set_to} in requirements.txt")


def build_mlops_wheel(deployment_folder: str) -> str:
    """Build akerbp.mlops wheel in a virtual environment

    Warning:
        This function assumes that the current version of the mlops package
        is a cloned copy of the mlops repo. It will not work if the package
        is installed from PyPI!

    Args:
        venv_dir (str): path to deployment folder where the wheel should be copied

    Returns:
        str: the wheel's filename.
    """
    # Get root of mlops repo
    root = str(Path(akerbp_mlops.__file__).parents[3])
    # Ensure the root contains pyproject.toml
    if not os.path.isfile(root + "/pyproject.toml"):
        raise DeploymentError("Failed to find pyproject.toml in root")
    # Build wheel
    logger.info("Building akerbp.mlops wheel")
    subprocess_wrapper([sys.executable, "-m", "build"], cwd=root)
    # Search for wheel name
    for file in os.listdir(root + "/dist"):
        if file.endswith(".whl"):
            wheel_name = file
            break
    else:
        # Remove local build artifacts
        logger.info("Removing local build artifacts")
        shutil.rmtree(root + "/dist")
        raise DeploymentError("Failed to build akerbp.mlops wheel")

    # Copy wheel to venv
    source = root + "/dist/" + wheel_name
    logger.info("Copying %s to virtual environment", source)
    shutil.copy(source, deployment_folder)

    # Remove local build artifacts
    logger.info("Removing local build artifacts")
    shutil.rmtree(root + "/dist")

    return str(Path(wheel_name).resolve().name)


def to_folder(path: Path, folder_path: Path) -> None:
    """
    Copy folders, files or package data to a given folder.
    Note that if target exists it will be overwritten.

    Args:
        path: supported formats
            - file/folder path (Path): e,g, Path("my/folder")
            - module file (tuple/list): e.g. ("my.module", "my_file"). Module
            path has to be a string, but file name can be a Path object.
        folder_path (Path): folder to copy to
    """
    if isinstance(path, (tuple, list)):
        module_path, file = path
        file = str(file)
        if importlib_resources.is_resource(module_path, file):
            with importlib_resources.path(module_path, file) as file_path:
                shutil.copy(file_path, folder_path)
        else:
            raise ValueError(f"Didn't find {path[1]} in {path[0]}")
    elif path.is_dir():
        shutil.copytree(
            path,
            folder_path / path,
            dirs_exist_ok=True,
            ignore=shutil.ignore_patterns("__pycache__", ".pytest_cache"),
        )
    elif path.is_file():
        shutil.copy(path, folder_path)
    else:
        raise ValueError(f"{path} should be a file, folder or package resource")


def get_deployment_folder_content(deployment_folder: Path) -> Dict:
    """Logs the content of the deployment folder as a dictionary with keys being
    directory/subdirectory, and values the corresponding content

    Args:
        deployment_folder (Path): path to deployment folder

    Returns:
        Dict: content of deployment folder
    """
    dirwalk = os.walk(deployment_folder)
    content = {}
    for root, dirs, files in dirwalk:
        if "__pycache__" in root.split("/"):
            continue
        if ".pytest_cache" in root.split("/"):
            continue
        dirs = [d for d in dirs if d not in ["__pycache__", ".pytest_cache"]]
        content[root] = files
        if len(dirs) > 0:
            content[root].extend(dirs)

            for subdir in dirs:
                content[os.path.join(root, subdir)] = files
        else:
            content[root] = files

    return content


def copy_to_deployment_folder(lst: Dict, deployment_folder: Path) -> None:
    """
    Copy a list of files/folders to a deployment folder

    Args:
        lst (dict): key is the nickname of the file/folder (used for
        logger) and the value is the path (see `to_folder` for supported
        formats)
        deployment_folder (Path): Path object for the deployment folder

    """
    for k, v in lst.items():
        if v:
            logger.debug(f"{k} => deployment folder")
            to_folder(v, deployment_folder)
        else:
            logger.warning(f"{k} has no value")


def update_pip(venv_dir: str, **kwargs) -> None:
    is_unix_os = kwargs.get("is_unix_os", True)
    setup_venv = kwargs.get("setup_venv", True)
    if setup_venv:
        if is_unix_os:
            sys.executable = os.path.join(venv_dir, "bin", "python")
            c = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
            ]
        else:
            sys.executable = os.path.join(venv_dir, "Scripts", "python")
            c = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
            ]
    else:
        c = ["python", "-m", "pip", "install", "--upgrade", "pip"]
    logger.info("Updating pip")
    subprocess_wrapper(c)


def install_requirements(req_file: str, venv_dir: str, **kwargs) -> None:
    """install model requirements

    Args:
        req_file (str): path to requirements file
    """
    with_deps = kwargs.get("with_deps", True)
    setup_venv = kwargs.get("setup_venv", False)
    is_unix_os = is_unix()
    update_pip(
        venv_dir=venv_dir,
        is_unix_os=is_unix_os,
        setup_venv=setup_venv,
    )
    logger.info(f"Installing python requirement file {req_file}:")
    # Cat file contents to log
    with open(req_file, "r") as f:
        logger.info(f.read())
    if with_deps:
        if setup_venv:
            if is_unix_os:
                sys.executable = os.path.join(venv_dir, "bin", "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-r",
                    os.path.abspath(req_file),
                ]
            else:
                sys.executable = os.path.join(venv_dir, "Scripts", "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-r",
                    os.path.abspath(req_file),
                ]
        else:
            c = ["pip", "install", "-r", os.path.abspath(req_file)]
    else:
        if setup_venv:
            if is_unix_os:
                sys.executable = os.path.join(venv_dir, "bin", "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--no-deps",
                    "-r",
                    os.path.abspath(req_file),
                ]
            else:
                sys.executable = os.path.join(venv_dir, "Scripts", "python")
                c = [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-r",
                    os.path.abspath(req_file),
                ]
        else:
            c = ["pip", "install", "--no-deps", "-r", os.path.abspath(req_file)]
    subprocess_wrapper(c, retry=3)


def create_venv(venv_name: str) -> str:
    venv_dir = os.path.join(os.getcwd(), venv_name)
    logger.info(f"Creating virtual environment {venv_name}")
    venv.create(venv_dir, with_pip=True)
    if os.path.isdir(venv_dir):
        logger.info(f"Successfully created virtual environment {venv_name}")
    else:
        raise VirtualEnvironmentError("Virtual environment was not created")
    return venv_dir


def delete_venv(venv_name: str) -> None:
    logger.info(f"Deleting virtual environment {venv_name}")
    subprocess_wrapper(["rm", "-rf", os.path.abspath(venv_name)])
    if not os.path.isdir(os.path.abspath(venv_name)):
        logger.info(f"Virtual environment {venv_name} sucsessfully deleted")
    else:
        raise VirtualEnvironmentError("Virtual environment not deleted")


def set_up_requirements(c: ServiceSettings, **kwargs) -> str:
    """
    Set up a "requirements.txt" file at the top of the deployment folder
    (assumed to be the current directory), update config and install
    dependencies (unless in dev)

    Args:
        c (ServiceSettings): service settings a specified in the config file

    Keyword Args:
        install (bool): Whether to install the dependencies, defaults to True
    """
    logger.info("Create requirement file")
    install_reqs = kwargs.get("install", True)
    with_deps = kwargs.get("with_deps", True)
    venv_name = kwargs.get("venv_name", "mlops-venv")
    setup_venv = kwargs.get("setup_venv", False)
    unit_testing = kwargs.get("unit_testing", False)
    if setup_venv:
        venv_dir = create_venv(venv_name=venv_name)
    else:
        venv_dir = str(os.getcwd())

    if not unit_testing:
        shutil.copyfile(c.req_file, "requirements.txt")
        c.req_file = Path("requirements.txt").resolve()

    if env != "dev" or install_reqs:
        install_requirements(
            c.req_file, venv_dir=venv_dir, with_deps=with_deps, setup_venv=setup_venv
        )
    else:
        logger.info("Skipping installation of requirements.txt")

    return venv_dir  # type: ignore


def deployment_folder_path(model: str) -> Path:
    """Generate path to deployment folder, which is on the form "mlops_<model>"

    Args:
        model (str): model name

    Returns:
        Path: path to the deployment folder
    """
    return Path(f"mlops_{model}")


def rm_deployment_folder(model: str) -> None:
    logger.info("Deleting deployment folder")
    deployment_folder = deployment_folder_path(model)
    if deployment_folder.exists():
        shutil.rmtree(deployment_folder)


def run_tests(c: ServiceSettings, **kwargs) -> Any:
    """Helper function that runs unit tests and returns a test payload
    if run during deployment

    Args:
        c (ServiceSettings): Service settings object containing model settings

    Keyword Args:
        setup_venv (bool): whether to setup an ephemeral virtual environment.
            Defaults to False
    Raises:
        TestError: If unit tests are failing

    Returns:
        Union[Dict[str, Any], None]: Return test payload if test file is specified in 'mlops_settings.yaml'
    """
    if c.test_file:
        deploy = kwargs.get("deploy", True)
        setup_venv = kwargs.get("setup_venv", False)
        executable = None
        if setup_venv:
            logger.info("Setting up ephemeral virtual environment")
            venv_dir = set_up_requirements(
                c,
                install=True,
                setup_venv=setup_venv,
            )
            if is_unix():
                executable = os.path.join(venv_dir, "bin", "python")
            else:
                executable = os.path.join(venv_dir, "Scripts", "python")
        else:
            set_up_requirements(c, install=True)
        test_command = [
            sys.executable if executable is None else executable,
            "-m",
            "akerbp.mlops.services.test_service",
        ]
        logger.info(f"Running tests for model {c.model_name}")
        failed_test = False
        try:
            output = subprocess.check_output(test_command, encoding="UTF-8")
        except subprocess.CalledProcessError as e:
            output = e.output
            failed_test = True
        model_input = None
        for log_line in output.splitlines():  # type: ignore
            log_line = log_line.strip()
            if log_line == "":
                pass
            elif log_line == "warnings.warn(":
                pass
            elif log_line.startswith('{"input": ['):
                # Extract payload for downstream testing of deployed model
                model_input = json.loads(log_line)
                if deploy:
                    logger.info(
                        "Payload for downstream testing of deployed model obtained"
                    )
            else:
                # Remove the date and time from piped log
                log_line = re.sub(
                    "\d{1,2}/\d{1,2}/\d{4} \d{1,2}:\d{1,2}:\d{1,2} -",
                    "",
                    log_line,
                )
                logger.info(log_line.strip())
        if failed_test:
            raise TestError("Unit tests failed :( See the above traceback")
        if model_input is None and deploy:
            raise TestError(
                "Test was not able to extract the payload for downstream testing of deployed model"
            )
        if setup_venv:
            delete_venv(venv_name=venv_dir.split("/")[-1])
            logger.info("Ephemeral virtual environment deleted")
        logger.info("Unit tests passed :)")

        return model_input
    else:
        logger.warning(
            "No test file specified in 'mlops_settings.yaml', skipping tests"
        )
        return {}


def do_deploy(
    c: ServiceSettings,
    env: str,
    service_name: str,
    function_name: str,
    deployment_folder: str,
    platform_methods: Dict = platform_methods,
    **kwargs,
) -> None:
    # Run unit tests and get test payload before deploying
    logger.info(
        f"Running tests for model {c.human_friendly_model_name} before deploying to {env}"
    )
    setup_venv = kwargs.pop("setup_venv", False)
    test_payload = run_tests(c, setup_venv=setup_venv)

    # Log deployment folder content - at this point we are already inside the deployment folder
    deployment_folder_content = get_deployment_folder_content(
        deployment_folder=Path(".")
    )
    logger.info(
        f"Deployment folder '{deployment_folder}' now contains the following: {deployment_folder_content}"
    )

    platform = c.platform
    deploy_function, _, test_function = platform_methods[platform]
    logger.info(f"Starting deployment of model {c.model_name} to {env}")
    if platform == "cdf":
        latest_artifact_version = cdf.get_latest_artifact_version(
            external_id=function_name
        )
        if c.artifact_version is None:
            logger.info(
                f"Latest artifact version in {env} is {latest_artifact_version}"
            )
            artifact_version = latest_artifact_version
        else:
            artifact_version = c.artifact_version

        external_id = function_name + "-" + str(artifact_version)
    elif platform == "gc":
        external_id = function_name  # Revisions are automatically created by GC
    else:
        raise NotImplementedError(
            f"Functionality for deploying models to platform {platform} is not implemented!"
        )

    logger.info(
        f"Deploying function {c.human_friendly_model_name} with external id {external_id} to {platform}"
    )
    try:
        deploy_function(
            c.human_friendly_model_name, external_id, info=c.info[service_name]
        )
    except Exception as e:
        raise DeploymentError(f"Deployment failed with message: \n{str(e)}") from e
    if c.test_file:
        logger.info(f"Make a test call to function with external id {external_id}")
        try:
            test_function(external_id, test_payload)
        except Exception as e:
            raise TestError(
                f"Test of deployed model failed with message: \n{str(e)}"
            ) from e
    else:
        logger.warning(
            f"No test file was set up. End-to-end test skipped for function {external_id}"
        )

    # Redeploy latest function with a predictable external id (model-service-env)
    if artifact_version == latest_artifact_version and platform == "cdf":
        logger.info(
            f"Redeploying latest model {c.human_friendly_model_name} with predictable external id {function_name} to {platform}"
        )
        redeploy_model_with_predictable_external_id(
            c,
            external_id=external_id,
            predictable_external_id=function_name,
            test_payload=test_payload,
        )

        # Create a schedule for keeping the latest function warm in prod
        if env == "prod":
            logger.info(
                f"Creating a schedule for keeping the function {function_name} warm on weekdays during extended working hours"
            )
            cdf.setup_schedule_for_latest_model_in_prod(
                predictable_external_id=function_name,
            )
    else:
        logger.info(
            f"Deployment is based on an old artifact version ({artifact_version}/{latest_artifact_version})."
        )
        logger.info("Skipping redeployment with a predictable external id.")
        if env == "prod" and platform == "cdf":
            logger.info(
                f"Will not setup a schedule for keeping this version in {env} warm"
            )
    if platform == "cdf":
        logger.info("Initiating garbage collection of old model versions")
        cdf.garbage_collection(
            c,
            function_name,
            env,
            remove_artifacts=c.model_name == "mlopsdemo",
        )


def redeploy_model_with_predictable_external_id(
    c: ServiceSettings,
    external_id: str,
    predictable_external_id: str,
    test_payload: Any,
    platform_methods: Dict = platform_methods,
) -> None:
    _, redeploy_function, test_function = platform_methods[c.platform]
    (
        name,
        file_id,
        description,
        metadata,
        owner,
    ) = cdf.get_arguments_for_redeploying_latest_model_version(external_id=external_id)

    try:
        redeploy_function(
            name,
            predictable_external_id,
            file_id,
            description,
            metadata,
            owner,
        )
    except Exception as e:
        raise DeploymentError(
            f"Redeployment of latest model failed with message:  \n{str(e)}"
        ) from e

    if c.test_file:
        try:
            test_function(predictable_external_id, test_payload)
        except Exception as e:
            raise TestError(
                f"Testing the newly redeployed latest model failed with message: \n{str(e)}"
            ) from e
    else:
        logger.warning("No test file was specified in the settings, skipping tests")


def create_temporary_copy(path: Path) -> Path:
    logger.info("Creating a temporary copy of %s", path)
    _, temp_path = tempfile.mkstemp()
    shutil.copy2(path, temp_path)
    return Path(temp_path).resolve()



import os
from subprocess import check_output


def get_current_version() -> str:
    """Get the current version of the package based on the latest release tag in git."""
    version = (
        check_output(["git", "describe", "--tags", "--abbrev=0"])
        .decode("ascii")
        .strip()
    )
    return version


def increment_version(
    version_to_test: str = "",
    commit_msg: str = "",
) -> str:
    """Increment version number by specifying type of update in the commit message.
    The function extracts the current version from a text file, examines the latest commit
    and increment the version number based on the content of the commit message and the environment.
    The updated version number is then updated in the text file keeping track of the current
    version of the package.

    The version number is tagged as a pre-release if the package is deployed in a test or dev environment.

    We increment the package version according to the following rules based on the commit message:
        - If commit message includes the word "major", increment MAJOR number
        - If commit message includes the word "minor", increment MINOR number
        - If commit message includes neither "major" nor "minor", increment PATCH number
        - Do not increment MAJOR, MINOR nor PATCH if the commit message includes the word "pre-release",
            only increment the pre-release number in this case to avoid unnecessary bumps.
    Returns:
        str: new version number
    """
    ENV = os.environ["MODEL_ENV"]
    if version_to_test:
        version = version_to_test
        commit_msg = commit_msg.strip().lower()
    else:
        version = get_current_version()
        commit_msg = (
            check_output(["git", "log", "-1", "--pretty=%B"]).decode().strip().lower()
        )
    if ENV == "prod":
        return version.split("a")[0]
    else:
        # Increment the package version based on the commit message in dev and test, and tag as pre-release
        major, minor, patch = version.split(".")
        if "a" in patch:
            is_pre_release = True
            patch, pre_release_version = patch.split("a")
        else:
            is_pre_release = False
            pre_release_version = None
        words_in_commit_msg = commit_msg.split()

        # Remove special characters from commit message
        for i, word in enumerate(words_in_commit_msg):
            alphanumeric = [char for char in word if char.isalnum()]
            word = "".join(alphanumeric)
            words_in_commit_msg[i] = word
        # Increment version based on commit message; Avoid incrementing if we specify to increment a pre-release
        if "prerelease" not in words_in_commit_msg:
            if "major" in words_in_commit_msg:
                major = str(int(major) + 1)
                minor, patch = "0", "0"
            elif "minor" in words_in_commit_msg:
                minor = str(int(minor) + 1)
                patch = "0"
            elif "patch" in words_in_commit_msg:
                patch = str(int(patch) + 1)
            else:
                raise ValueError(
                    "Commit message must include either major, minor, or patch keyword if prerelease is not in commit message!"
                )

        new_version = f"{major}.{minor}.{patch}"
        # Tag version number with pre-release, increment pre-release number
        new_version += "a"
        if is_pre_release and pre_release_version is not None:
            pre_release_version_ = int(pre_release_version) + 1
            new_version += str(pre_release_version_)
        else:
            new_version += "0"

        return new_version


if __name__ == "__main__":
    print(increment_version())
